# utilities.py
# -*- coding: utf-8 -*-
#
# This file is part of LilyPond, the GNU music typesetter.
#
# Copyright (C) 2016--2026 John Gourlay <john@weathervanefarm.net>
#
# LilyPond is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# LilyPond is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with LilyPond.  If not, see <http://www.gnu.org/licenses/>.


import re


def escape_ly_output_string(input_string):
    return_string = input_string
    needs_quotes = not re.match("^[a-zA-ZäöüÜÄÖßñ]+$", return_string)
    if needs_quotes:
        return_string = ('"'
                         + return_string.replace('\\', '\\\\')
                                        .replace('"', r'\"')
                         + '"')
    return return_string


def interpret_alter_element(alter_elm):
    alter = 0
    if alter_elm:
        val = eval(alter_elm.get_text())
        if type(val) in (int, float):
            alter = val
    return alter


def musicxml_duration_to_log(dur):
    return {'1024th': 10,
            '512th': 9,
            '256th': 8,
            '128th': 7,
            '64th': 6,
            '32nd': 5,
            '16th': 4,
            'eighth': 3,
            'quarter': 2,
            'half': 1,
            'whole': 0,
            'breve': -1,
            'longa': -2,
            'long': -2}.get(dur, 0)


def round_to_two_digits(val):
    return round(val * 100) / 100


# Examples:
#
#   r'"foo  \"bar  baz\"  urgh"'  ->  [r'"foo  \"bar  baz\"  urgh"']
#   r'foo  "bar  baz"  urgh'      ->  ['foo', '"bar  baz"', 'urgh']
#
# We only split at ASCII whitespace characters, mainly to preserve the
# non-breakable space character (U+0080).
def split_string_and_preserve_doublequoted_substrings(value):
    return re.findall(r'(?asx) (?: " .*? [^\\] " | \S )+', value)


def musicxml_sound_to_lilypond_midi_instrument(sound):
    sounds = {
        "brass.french-horn": 'french horn',
        "brass.group": 'brass section',
        "brass.group.synth": 'synthbrass 1',
        "brass.trombone": 'trombone',
        "brass.trombone.alto": 'trombone',
        "brass.trombone.bass": 'trombone',
        "brass.trombone.contrabass": 'trombone',
        "brass.trombone.tenor": 'trombone',
        "brass.trumpet": 'trumpet',
        "brass.trumpet.baroque": 'trumpet',
        "brass.trumpet.bass": 'trumpet',
        "brass.trumpet.bflat": 'trumpet',
        "brass.trumpet.c": 'trumpet',
        "brass.trumpet.d": 'trumpet',
        "brass.trumpet.piccolo": 'trumpet',
        "brass.trumpet.pocket": 'trumpet',
        "brass.trumpet.slide": 'trumpet',
        "brass.trumpet.tenor": 'trumpet',
        "brass.tuba": 'tuba',
        "brass.tuba.bass": 'tuba',
        "brass.tuba.subcontrabass": 'tuba',
        "brass.wagner-tuba": 'french horn',
        "drum.timpani": 'timpani',
        "drum.tom-tom": 'melodic tom',
        "drum.tom-tom.synth": 'synth drum',
        "effect.applause": 'applause',
        "effect.bass-string-slap": 'slap bass 1',
        "effect.bird": 'bird tweet',
        "effect.bird.tweet": 'bird tweet',
        "effect.breath": 'breath noise',
        "effect.guitar-fret": 'guitar fret noise',
        "effect.gunshot": 'gunshot',
        "effect.helicopter": 'helicopter',
        "effect.metronome-click": 'woodblock',
        "effect.rain": 'fx1 (rain)',
        "effect.seashore": 'seashore',
        "effect.telephone-ring": 'telephone ring',
        "keyboard.accordion": 'accordion',
        "keyboard.bandoneon": 'accordion',
        "keyboard.celesta": 'celesta',
        "keyboard.clavichord": 'clav',
        "keyboard.concertina": 'concertina',
        "keyboard.harpsichord": 'harpsichord',
        "keyboard.ondes-martenot": 'ocarina',
        "keyboard.organ": 'church organ',
        "keyboard.organ.drawbar": 'drawbar organ',
        "keyboard.organ.percussive": 'percussive organ',
        "keyboard.organ.pipe": 'church organ',
        "keyboard.organ.reed": 'reed organ',
        "keyboard.piano": 'acoustic grand',
        "keyboard.piano.electric": 'electric piano 1',
        "keyboard.piano.grand": 'acoustic grand',
        "keyboard.piano.honky-tonk": 'honky-tonk',
        "metal.bells.agogo": 'agogo',
        "metal.bells.tinklebell": 'tinkle bell',
        "metal.cymbal.reverse": 'reverse cymbal',
        "pitched-percussion.glockenspiel": 'glockenspiel',
        "pitched-percussion.glockenspiel.alto": 'glockenspiel',
        "pitched-percussion.glockenspiel.soprano": 'glockenspiel',
        "pitched-percussion.hammer-dulcimer": 'dulcimer',
        "pitched-percussion.kalimba": 'kalimba',
        "pitched-percussion.marimba": 'marimba',
        "pitched-percussion.marimba.bass": 'marimba',
        "pitched-percussion.music-box": 'music box',
        "pitched-percussion.tubular-bells": 'tubular bells',
        "pitched-percussion.vibraphone": 'vibraphone',
        "pitched-percussion.xylophone": 'xylophone',
        "pitched-percussion.xylophone.alto": 'xylophone',
        "pitched-percussion.xylophone.bass": 'xylophone',
        "pitched-percussion.xylophone.soprano": 'xylophone',
        "pitched-percussion.xylorimba": 'xylophone',
        "pluck.banjo": 'banjo',
        "pluck.banjo.tenor": 'banjo',
        "pluck.bass": 'acoustic bass',
        "pluck.bass.acoustic": 'acoustic bass',
        "pluck.bass.electric": 'electric bass',
        "pluck.bass.fretless": 'fretless bass',
        "pluck.bass.synth": 'synth bass 1',
        "pluck.dulcimer": 'dulcimer',
        "pluck.guitar": 'acoustic guitar (nylon)',
        "pluck.guitar.acoustic": 'acoustic guitar (nylon)',
        "pluck.guitar.electric": 'electric guitar (jazz)',
        "pluck.guitar.nylon-string": 'acoustic guitar (nylon)',
        "pluck.guitar.steel-string": 'acoustic guitar (steel)',
        "pluck.harp": 'orchestral harp',
        "pluck.lute": 'acoustic guitar (nylon)',
        "pluck.shamisen": 'shamisen',
        "pluck.sitar": 'sitar',
        "strings.cello": 'cello',
        "strings.cello.piccolo": 'cello',
        "strings.contrabass": 'contrabass',
        "strings.fiddle": 'fiddle',
        "strings.group.synth": 'synth strings 1',
        "strings.viola": 'viola',
        "strings.violin": 'violin',
        "synth.effects.atmosphere": 'fx 4 (atmosphere)',
        "synth.effects.brightness": 'fx 5 (brightness)',
        "synth.effects.crystal": 'fx 3 (crystal)',
        "synth.effects.echoes": 'fx 7 echoes',
        "synth.effects.goblins": 'fx 6 goblins',
        "synth.effects.rain": 'fx 1 rain',
        "synth.effects.sci-fi": 'fx 8 sci-fi',
        "synth.effects.soundtrack": 'fx 2 (soundtrack)',
        "synth.pad.bowed": 'pad 5 bowed',
        "synth.pad.choir": 'pad 4 choir',
        "synth.pad.halo": 'pad 7 halo',
        "synth.pad.metallic": 'pad 6 metallic',
        "synth.pad.polysynth": 'pad 3 polysynth',
        "synth.pad.sweep": 'pad 8 sweep',
        "synth.pad.warm": 'pad 2 warm',
        "synth.tone.sawtooth": 'lead 1 (square)',
        "synth.tone.square": 'lead 2 (sawtooth)',
        "voice.aa": 'choir aahs',
        "voice.alto": 'choir aahs',
        "voice.aw": 'choir aahs',
        "voice.baritone": 'choir aahs',
        "voice.bass": 'choir aahs',
        "voice.child": 'choir aahs',
        "voice.countertenor": 'choir aahs',
        "voice.doo": 'choir aahs',
        "voice.ee": 'choir aahs',
        "voice.female": 'choir aahs',
        "voice.kazoo": 'choir aahs',
        "voice.male": 'choir aahs',
        "voice.mezzo-soprano": 'choir aahs',
        "voice.mm": 'choir aahs',
        "voice.oo": 'voice oohs',
        "voice.soprano": 'choir aahs',
        "voice.synth": 'synth voice',
        "wind.flutes.blown-bottle": 'blown bottle',
        "wind.flutes.calliope": 'lead 3 (calliope)',
        "wind.flutes.flute": 'flute',
        "wind.flutes.flute.alto": 'flute',
        "wind.flutes.flute.bass": 'flute',
        "wind.flutes.flute.contra-alto": 'flute',
        "wind.flutes.flute.contrabass": 'flute',
        "wind.flutes.flute.double-contrabass": 'flute',
        "wind.flutes.flute.piccolo": 'flute',
        "wind.flutes.flute.subcontrabass": 'flute',
        "wind.flutes.ocarina": 'ocarina',
        "wind.flutes.recorder": 'recorder',
        "wind.flutes.recorder.alto": 'recorder',
        "wind.flutes.recorder.bass": 'recorder',
        "wind.flutes.recorder.contrabass": 'recorder',
        "wind.flutes.recorder.descant": 'recorder',
        "wind.flutes.recorder.garklein": 'recorder',
        "wind.flutes.recorder.great-bass": 'recorder',
        "wind.flutes.recorder.sopranino": 'recorder',
        "wind.flutes.recorder.soprano": 'recorder',
        "wind.flutes.recorder.tenor": 'recorder',
        "wind.flutes.shakuhachi": 'shakuhachi',
        "wind.flutes.whistle": 'whistle',
        "wind.flutes.whistle.alto": 'whistle',
        "wind.pipes.bagpipes": 'bagpipe',
        "wind.reed.basset-horn": 'clarinet',
        "wind.reed.bassoon": 'bassoon',
        "wind.reed.clarinet": 'clarinet',
        "wind.reed.clarinet.a": 'clarinet',
        "wind.reed.clarinet.alto": 'clarinet',
        "wind.reed.clarinet.bass": 'clarinet',
        "wind.reed.clarinet.basset": 'clarinet',
        "wind.reed.clarinet.bflat": 'clarinet',
        "wind.reed.clarinet.contra-alto": 'clarinet',
        "wind.reed.clarinet.contrabass": 'clarinet',
        "wind.reed.clarinet.eflat": 'clarinet',
        "wind.reed.clarinet.piccolo.aflat": 'clarinet',
        "wind.reed.contrabass": 'contrabass',
        "wind.reed.contrabassoon": 'bassoon',
        "wind.reed.english-horn": 'oboe',
        "wind.reed.harmonica": 'harmonica',
        "wind.reed.harmonica.bass": 'harmonica',
        "wind.reed.oboe": 'oboe',
        "wind.reed.oboe.bass": 'oboe',
        "wind.reed.oboe.piccolo": 'oboe',
        "wind.reed.oboe-da-caccia": 'oboe',
        "wind.reed.oboe-damore": 'oboe',
        "wind.reed.saxophone": 'alto sax',
        "wind.reed.saxophone.alto": 'alto sax',
        "wind.reed.saxophone.baritone": 'baritone sax',
        "wind.reed.saxophone.bass": 'baritone sax',
        "wind.reed.saxophone.contrabass": 'baritone sax',
        "wind.reed.saxophone.melody": 'soprano sax',
        "wind.reed.saxophone.mezzo-soprano": 'soprano sax',
        "wind.reed.saxophone.sopranino": 'soprano sax',
        "wind.reed.saxophone.sopranissimo": 'soprano sax',
        "wind.reed.saxophone.soprano": 'soprano, sax',
        "wind.reed.saxophone.subcontrabass": 'baritone sax',
        "wind.reed.saxophone.tenor": 'tenor sax',
        "wind.reed.shenai": 'shanai',
        "wood.temple-block": 'wood block',
        "wood.wood-block": 'wood block',
    }
    return sounds.get(sound, 'acoustic grand')



# -*- coding: utf-8 -*-
#
# This file is part of LilyPond, the GNU music typesetter.
#
# Copyright (C) 2024 Werner Lemberg <wl@gnu.org>
#
# LilyPond is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# LilyPond is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with LilyPond.  If not, see <http://www.gnu.org/licenses/>.


import musicexp

# Use a global variable to store the settings needed inside a `\layout`
# block.  Whenever we need to change a setting or add/remove an engraver, we
# can access this variable and add the corresponding settings.
layout_information = musicexp.Layout()

# Use a global variable to store the settings needed inside a `\paper`
# block.
paper = musicexp.Paper()



# book_latex.py
# -*- coding: utf-8 -*-
#
# This file is part of LilyPond, the GNU music typesetter.
#
# Copyright (C) 2020--2026 Han-Wen Nienhuys <hanwen@xs4all.nl>,
# Copyright (C) 2010--2026 Reinhold Kainhofer <reinhold@kainhofer.com>
#
# LilyPond is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# LilyPond is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with LilyPond.  If not, see <http://www.gnu.org/licenses/>.


import os
import re
import subprocess
import sys
import tempfile

import book_base
import book_snippets
import lilylib as ly


# Recognize special sequences in the input.
#
#   (?P<name>regex) -- Assign result of REGEX to NAME.
#   *?              -- Match non-greedily.
#   (?!...)         -- Match if `...' doesn't match next (without
#                      consuming the string).
#   (?(name)yes|no) -- If group NAME exists, match YES pattern;
#                      otherwise match NO pattern
#
#   (?m) -- Multiline regex: Make ^ and $ match at each line.
#   (?s) -- Make the dot match all characters including newline.
#   (?x) -- Ignore whitespace in patterns.
#
# See book_base.BookOutputFormat for possible keys.
Latex_snippet_res = {
    'include':
    r'''(?smx)
          ^[^%\n]*?
          (?P<match>
          \\input\s*{
           (?P<filename>\S+?)
          })''',

    'lilypond':
    r'''(?smx)
          ^ [^%\n]*?
          (?P<match>
            \\lilypond
            \s*
            ( \[ \s* (?P<options> [^\[\]]*? ) \s* \] )?
            \s*
            { (?P<code>''' + ly.brace_matcher(10) + r''') \s* }
          )''',

    # Accept both
    #   \begin[options]{lilypond}
    # and
    #   \begin{lilypond}[options]
    'lilypond_block':
    r'''(?smx)
          ^ [^%\n]*?
          (?P<match>
            \\begin
            \s*
            (?P<env> {lilypond} \s* )?
            ( \[ \s* (?P<options> [^\[\]]*? ) \s* \] )?
            (?(env) | \s* {lilypond} )
            (?P<code> .*? ) \s*
            ^ [^%\n]*?
            \\end \s* {lilypond}
          )''',

    'lilypond_file':
    r'''(?smx)
          ^ [^%\n]*?
          (?P<match>
            \\lilypondfile
            \s*
            ( \[ \s* (?P<options> [^\[\]]*? ) \s* \] )?
            \s*
            { (?P<filename> \S+? ) }
          )''',

    'musicxml_file':
    r'''(?smx)
          ^ [^%\n]*?
          (?P<match>
            \\musicxmlfile
            \s*
            ( \[ \s* (?P<options> [^\[\]]*? ) \s* \] )?
            \s*
            { (?P<filename> \S+? ) }
          )''',

    'singleline_comment':
    r'''(?mx)
          ^.*?
          (?P<match>
           (?P<code>
           %.*$\n+))''',

    'verb':
    r'''(?mx)
          ^[^%\n]*?
          (?P<match>
           (?P<code>
           \\verb(?P<del>.)
            .*?
           (?P=del)))''',

    'verbatim':
    r'''(?msx)
          ^[^%\n]*?
          (?P<match>
           (?P<code>
           \\begin\s*{verbatim}
            .*?
           \\end\s*{verbatim}))''',

    'lilypondversion':
    r'''(?smx)
          (?P<match>
          \\lilypondversion)[^a-zA-Z]''',
}

Latex_output = {
    book_snippets.FILTER: r'''\begin{lilypond}[%(options)s]
%(code)s
\end{lilypond}''',

    book_snippets.OUTPUT: r'''{%%
\parindent 0pt
\noindent
\ifx\preLilyPondExample \undefined
\else
  \expandafter\preLilyPondExample
\fi
\input{%(base)s-systems.tex}%%
\ifx\postLilyPondExample \undefined
\else
  \expandafter\postLilyPondExample
\fi
}''',

    book_snippets.INLINEOUTPUT: r'''%%
\ifx\preLilyPondExample \undefined
\else
  \expandafter\preLilyPondExample
\fi
\raisebox{%(vshift)s\height}{\input{%(base)s-systems.tex}}%%
\ifx\postLilyPondExample \undefined
\else
  \expandafter\postLilyPondExample
\fi
{}''',

    book_snippets.PRINTFILENAME: r'''\texttt{%(filename)s}
\linebreak
''',

    book_snippets.QUOTE: r'''\begin{quote}
%(str)s
\end{quote}''',

    book_snippets.VERBATIM: r'''\begin{verbatim}%(verb)s\end{verbatim}
''',

    book_snippets.VERSION: r'''%(program_version)s''',
}


###
# Retrieve dimensions from LaTeX
LATEX_INSPECTION_DOCUMENT = r'''
\nonstopmode
%(preamble)s
\begin{document}
\typeout{paperwidth=\the\paperwidth}
\typeout{paperheight=\the\paperheight}
\typeout{textwidth=\the\textwidth}
\typeout{columnsep=\the\columnsep}
\makeatletter\if@twocolumn\typeout{columns=2}\fi\makeatother
\end{document}
'''


def extract_latex_geometry_setting(setting, parameter_string):
    m = re.search(setting + '=([-0-9.]+)pt', parameter_string)
    if m:
        return float(m.group(1))


def format_pt(value):
    return '%.2f\\pt' % value


def get_latex_paper_geometry(source, global_options):
    m = re.search(r'''(?P<preamble>\\begin\s*{document})''', source)
    if m is None:
        ly.warning(_("cannot find \\begin{document} in LaTeX document"))
        return {}

    preamble = source[:m.start(0)]
    latex_document = LATEX_INSPECTION_DOCUMENT % {'preamble': preamble}

    (handle, tmpfile) = tempfile.mkstemp('.tex')
    tmpfileroot = os.path.splitext(tmpfile)[0]
    tmpfileroot = os.path.split(tmpfileroot)[1]
    auxfile = tmpfileroot + '.aux'
    logfile = tmpfileroot + '.log'

    tmp_handle = os.fdopen(handle, 'w')
    tmp_handle.write(latex_document)
    tmp_handle.close()

    ly.progress(_("Running `%s' on file `%s' to detect default page settings.\n")
             % (global_options.latex_program, tmpfile))
    cmd = '%s %s' % (global_options.latex_program, tmpfile)
    ly.debug_output("Executing: %s\n" % cmd)
    run_env = os.environ.copy()
    run_env['LC_ALL'] = 'C'
    run_env['TEXINPUTS'] = os.path.pathsep.join(
                             (global_options.input_dir,
                              run_env.get('TEXINPUTS', '')))

    # unknown why this is necessary
    universal_newlines = True
    if sys.platform == 'mingw32':
        universal_newlines = False
        # use os.system to avoid weird sleep() problems on
        # GUB's python 2.4.2 on mingw
        # make file to write to
        output_dir = tempfile.mkdtemp()
        output_filename = os.path.join(output_dir, 'output.txt')
        # call command
        cmd += " > %s" % output_filename
        oldtexinputs = os.environ.get('TEXINPUTS')
        os.environ['TEXINPUTS'] = run_env['TEXINPUTS']
        returncode = os.system(cmd)
        if oldtexinputs:
            os.environ['TEXINPUTS'] = oldtexinputs
        else:
            del os.environ['TEXINPUTS']
        parameter_string = open(output_filename, encoding="utf8").read()
        if returncode != 0:
            ly.warning(_("Unable to auto-detect default settings:\n"))
        # clean up
        os.remove(output_filename)
        os.rmdir(output_dir)
    else:
        proc = subprocess.Popen(cmd,
                                env=run_env,
                                universal_newlines=universal_newlines,
                                shell=True,
                                stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        (parameter_string, error_string) = proc.communicate()
        if proc.returncode != 0:
            if not error_string:
                error_string = (
                    "  LaTeX subprocess returned error code %s"
                    % proc.returncode)
            ly.warning(_("Unable to auto-detect default settings:\n%s")
                    % error_string)
    os.unlink(tmpfile)
    if os.path.exists(auxfile):
        os.unlink(auxfile)
    if os.path.exists(logfile):
        parameter_string = open(logfile, encoding="utf8").read()
        os.unlink(logfile)

    columns = 1
    m = re.search('columns=([0-9.]+)', parameter_string)
    if m:
        columns = int(m.group(1))

    paperwidth = extract_latex_geometry_setting('paperwidth',
                                                parameter_string)
    paperheight = extract_latex_geometry_setting('paperheight',
                                                 parameter_string)
    textwidth = extract_latex_geometry_setting('textwidth',
                                               parameter_string)
    columnsep = extract_latex_geometry_setting('columnsep',
                                               parameter_string)

    ly.debug_output('Detected values:')
    ly.debug_output('  paperwidth = %s' % paperwidth)
    ly.debug_output('  paperheight = %s' % paperheight)
    ly.debug_output('  textwidth = %s' % textwidth)
    ly.debug_output('  columns = %s' % columns)
    ly.debug_output('  columnsep = %s\n' % columnsep)

    if textwidth and columns > 1 and columnsep:
        textwidth = (textwidth + columnsep) / columns - columnsep
        ly.debug_output('Adjusted value:')
        ly.debug_output('  textwidth = %s' % textwidth)

    geometry = {}
    if paperwidth:
        geometry[book_snippets.PAPER_WIDTH] = format_pt(paperwidth)
    if paperheight:
        geometry[book_snippets.PAPER_HEIGHT] = format_pt(paperheight)
    if textwidth:
        geometry[book_snippets.LINE_WIDTH] = format_pt(textwidth)

    return geometry


def modify_preamble(chunk):
    s = chunk.replacement_text()
    if (re.search(r"\\begin *{document}", s)
            and not re.search("{graphic[sx]", s)):
        s = re.sub(r"\\begin{document}",
                     r"\\usepackage{graphics}" + '\n'
                     + r"\\begin{document}",
                     s)
        chunk.override_text = s


class BookLatexOutputFormat (book_base.BookOutputFormat):
    def __init__(self):
        book_base.BookOutputFormat.__init__(self)
        self.format = "latex"
        self.default_extension = ".tex"
        self.snippet_res = Latex_snippet_res
        self.output = Latex_output
        self.handled_extensions = ['.latex', '.lytex', '.tex']
        self.image_formats = "ps"
        self.snippet_option_separator = r'\s*,\s*'

    def process_options(self, global_options):
        self.process_options_pdfnotdefault(global_options)

    def get_paper_geometry(self, source):
        return get_latex_paper_geometry(source, self.global_options)

    def input_fullname(self, input_filename):
        # Use kpsewhich if available, otherwise fall back to the default:
        try:
            input_fullname = subprocess.run(['kpsewhich', input_filename],
                                            check=True,
                                            encoding='utf-8',
                                            stdout=subprocess.PIPE,
                                            universal_newlines=True).stdout

            input_fullname = input_fullname.strip("\n")

        except (subprocess.CalledProcessError, FileNotFoundError):
            input_fullname = book_base.BookOutputFormat.input_fullname(
                self,
                input_filename)

        return input_fullname

    def process_chunks(self, chunks):
        for c in chunks:
            if (c.is_plain() and
                    re.search(r"\\begin *{document}", c.replacement_text())):
                modify_preamble(c)
                break
        return chunks

    def snippet_output(self, basename, snippet):
        s = ''
        rep = snippet.get_replacements()
        rep['base'] = basename.replace('\\', '/')
        rep['filename'] = os.path.basename(snippet.filename).replace('\\', '/')
        rep['ext'] = snippet.ext
        if book_snippets.PRINTFILENAME in snippet.option_dict:
            s += self.output[book_snippets.PRINTFILENAME] % rep
        if book_snippets.VERBATIM in snippet.option_dict:
            rep['verb'] = snippet.verb_ly()
            s += self.output[book_snippets.VERBATIM] % rep

        if book_snippets.INLINE not in snippet.option_dict:
            s += self.output[book_snippets.OUTPUT] % rep
        else:
            rep['vshift'] = snippet.option_dict[book_snippets.INLINE]
            s += self.output[book_snippets.INLINEOUTPUT] % rep

        # todo: maintain breaks
        if 0:
            breaks = snippet.ly().count("\n")
            s += "".ljust(breaks, "\n").replace("\n", "%\n")

        if book_snippets.QUOTE in snippet.option_dict:
            s = self.output[book_snippets.QUOTE] % {'str': s}

        return s


book_base.register_format(BookLatexOutputFormat())



# -*- coding: utf-8 -*-
#
# This file is part of LilyPond, the GNU music typesetter.
#
# Copyright (C) 2016--2026 John Gourlay <john@weathervanefarm.net>
# Copyright (C) 2024--2026 Werner Lemberg <wl@gnu.org>
#
# LilyPond is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# LilyPond is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with LilyPond.  If not, see <http://www.gnu.org/licenses/>.


additional_definitions = {
    "accidental-marks": """\
% \\accs-ornament
% --------------
%
% Position the `above` and `below` markup lists as vertical stacks
% above and below markup `script`, respectively.  `above` and `below`
% are by default horizontally centered on `script` (controlled by the
% property `acc-dir`), printed at a size given by the property
% `acc-font-size` (default value -5), and a color given by the
% property `acc-color` (default is black).  Between all markups a
% vertical space given by the property `acc-padding` (default value
% 0.2) is inserted.  The `above` stack grows from bottom to top, while
% the `below` stack grows from top to bottom.
%
% If either `above` or `below` is an empty argument, or if `script` is
% set to `#f`, the corresponding argument is not printed.
%
% If an element of `above`, `script`, or `below` is a plain string,
% check the following.
%
% * If the string consists entirely of one or more characters from the
%   accidental set '♮♭♯𝄪𝄫' (not counting a possible enclosure), add
%   `\\number` internally as a prefix to use the Emmentaler font.
%
% * Otherwise the element is taken as a glyph name (again not counting
%   a possible enclosure), to be internally accessed with the
%   `\\musicglyph` command.
%
% * (Not for `script`.)  If an element of `above` or `below` is a
%   plain string and the first and last character is `(` and `)`,
%   respectively, the element gets enclosed in parentheses.  The same
%   holds for `[` and `]`, enclosing the element in brackets.  The
%   font size of the enclosure characters is controlled by the
%   property `enclosure-font-size` (default value -2).
%
% Example:
%
% ```
% \\markup \\override #'(acc-padding . 0.5)
%         \\accs-ornament { "♭" "(♯)" } "scripts.haydnturn" {}
% ```

#(define-markup-command (accs-ornament layout props
                                       above script below)
   (markup-list? markup? markup-list?)
   #:properties ((acc-font-size -5)
                 (acc-color black)
                 (acc-padding 0.3)
                 (acc-dir CENTER)
	         (enclosure-font-size -2))

   (define enclosure-regex
     ;; `(?(-3) ...)` is a 'conditional subpattern' that is only
     ;; considered for matching if the subpattern three opening
     ;; parentheses earlier (i.e., the first group) matches.
     (ly:make-regex "(?x) ^ ( [([] ) ?
                            ( [^]()[] + )
                            ( (?(-3) [])] ) ) $"))

   (define charset:accidentals (string->char-set "♮♭♯𝄪𝄫"))

   (define (musicglyph-or-number-markup arg)
     (if (string-every charset:accidentals arg)
         (make-number-markup arg)
         (make-musicglyph-markup arg)))

   (define (with-enclosure-markup arg)
     (let ((match (ly:regex-exec enclosure-regex arg)))
       (if match
           (let* (;; `left` is `#f` if there is no enclosure.
                  (left (ly:regex-match-substring match 1))
                  (left (if left
                            (make-normalsize-markup
                             (make-fontsize-markup
                              enclosure-font-size left))
                            #f))
                  (glyph (ly:regex-match-substring match 2))
                  ;; `right` is `""` if there is no enclosure.
                  (right (ly:regex-match-substring match 3))
                  (right (if (string-null? right)
                             #f
                             (make-normalsize-markup
                              (make-fontsize-markup
                               enclosure-font-size right)))))
             (make-concat-markup
              (list
               (or left "")
               (musicglyph-or-number-markup glyph)
               (or right ""))))
           #f)))

   (define (to-markup arg enclosure)
     (if (string? arg)
	 (if (string-null? arg)
             #f
             (if enclosure
                 (with-enclosure-markup arg)
                 (musicglyph-or-number-markup arg)))
         arg))

   (define (make-acc-stencil arg)
     (if arg
         (interpret-markup
          layout props
          (make-halign-markup
           acc-dir
           (make-with-color-markup
            acc-color
            (make-fontsize-markup
             acc-font-size arg))))
         #f))

   (let* ((above (map (lambda (x) (to-markup x #t)) above))
          (above-stils (map (lambda (x)
                              (make-acc-stencil x))
                            above))

          (script (to-markup script #f))
          (script-stil (if script
                           (interpret-markup layout props script)
                           empty-stencil))

          (below (map (lambda (x) (to-markup x #t)) below))
          (below-stils (map (lambda (x)
                              (make-acc-stencil x))
                            below))

          (script-stil (fold (lambda (elem previous)
                               (ly:stencil-combine-at-edge
                                previous Y UP elem acc-padding))
                             script-stil
                             above-stils))
          (script-stil (fold (lambda (elem previous)
                               (ly:stencil-combine-at-edge
                                previous Y DOWN elem acc-padding))
                             script-stil
                             below-stils)))
     script-stil))


% \\acc-ornament
% -------------
%
% This function behaves similar to `\\accs-ornament` with the
% difference that `above` and `below` are single markups, not markup
% lists.  If one of the arguments is an empty string, it is not
% printed.
%
% Example:
%
% ```
% \\markup \\acc-ornament "[♭]" "scripts.turn" ♯
% ```

#(define-markup-command (acc-ornament layout props
                                      above script below)
   (markup? markup? markup?)
   #:properties (accs-ornament-markup)
   (accs-ornament-markup layout props
                         (if (and (string? above) (string-null? above))
                             '()
                             (list above))
                         (if (and (string? script) (string-null? script))
                             #f
                             script)
                         (if (and (string? below) (string-null? below))
                             '()
                             (list below))))


% Some shorthands to make the usage of `\\acc-ornament` more
% comfortable for simple cases.


ornament =
#(define-music-function (above script below)
   (string? string? string?)
   #{
     \\tweak parent-alignment-X #CENTER
     \\tweak self-alignment-X #CENTER
     -\\markup \\acc-ornament #above #script #below
   #})


accTrill =
#(define-music-function (above)
   (string?)
   #{
     \\tweak parent-alignment-X #CENTER
     \\tweak self-alignment-X #CENTER
     -\\markup
        \\override #'((acc-dir . -1.3)
                     (acc-padding . -0.4))
        \\acc-ornament #above "scripts.trill" ""
   #})


#(define-markup-command (trill-acc-tweak layout props above)
   (string?)
   (interpret-markup
    layout props
    #{
      \\markup
        \\override #'((acc-dir . -1.3)
                     (acc-padding . -0.4))
        \\with-dimension-from #X
          \\with-true-dimension #X \\musicglyph "scripts.trill"
            \\acc-ornament
              #above
              \\with-true-dimension #X \\musicglyph "scripts.trill"
          ""
    #}))


% `color` and `size` are optional.  If `size` is present, `color` must
% be present, too.  `size` also sets the `enclosure-font-size`
% property (three magsteps larger).
trillTweak =
#(define-music-function (acc color size music)
   (string? (color? black) (number?) ly:music?)
   (let* ((override (list `(acc-color . ,color)))
          (override (if size
                        (cons `(enclosure-font-size . ,(+ size 3))
                         (cons `(acc-font-size . ,size) override))
                        override)))
     #{
       \\tweak bound-details.left.text
         \\markup \\override #override
                   \\trill-acc-tweak #acc
       #music
     #}))
""",

    "arpeggioXX": """\
% For cross-staff arpeggios.
arpeggioXX =
  \\once \\set PianoStaff.connectArpeggios = ##t
arpeggioArrowUpXX =
  \\once \\override PianoStaff.Arpeggio.arpeggio-direction = #UP
arpeggioArrowDownXX =
  \\once \\override PianoStaff.Arpeggio.arpeggio-direction = #DOWN
arpeggioBracketXX =
  \\once \\override PianoStaff.Arpeggio.stencil =
    #ly:chord-bracket::print
""",

    "arpeggioX": """\
% For cross-voice arpeggios.
arpeggioX =
  \\once \\set Staff.connectArpeggios = ##t
arpeggioArrowUpX =
  \\once \\override Staff.Arpeggio.arpeggio-direction = #UP
arpeggioArrowDownX =
  \\once \\override Staff.Arpeggio.arpeggio-direction = #DOWN
arpeggioBracketX =
  \\once \\override Staff.Arpeggio.stencil =
    #ly:chord-bracket::print
""",

    "cancel-after-key": """\
cancelAfterKey = {
  \\once \\set Staff.printKeyCancellation = ##t
  \\once \\override Score.BreakAlignment.break-align-orders =
  #(grob-transformer
    'break-align-orders
    (lambda (grob orig)
      (let* ((middle (vector-ref orig 1))
             (middle (delq 'key-signature middle))
             (middle (insert-before 'key-cancellation 'key-signature middle)))
        (vector
         (vector-ref orig 0)
         middle
         (vector-ref orig 2)))))
}
""",

    "cancel-before-barline": """\
cancelBeforeBarline = {
  \\once \\set Staff.printKeyCancellation = ##t
  \\once \\override Score.BreakAlignment.break-align-orders =
  #(grob-transformer
    'break-align-orders
    (lambda (grob orig)
      (let* ((middle (vector-ref orig 1))
             (middle (delq 'key-cancellation middle))
             (middle (insert-before 'staff-bar 'key-cancellation middle)))
        (vector
         (vector-ref orig 0)
         middle
         (vector-ref orig 2)))))
}
""",

    "crescendo": """\
Cresc = #(make-music 'CrescendoEvent 'span-direction START
                                     'span-type 'text)
""",

    "decrescendo": """\
Decresc = #(make-music 'DecrescendoEvent 'span-direction START
                                         'span-type 'text)
""",

    "fingering-substitution": """\
substFinger =
#(define-music-function (start left right)
   (markup? markup? markup?)
   #{
     \\tweak self-alignment-X #-0.6
     -\\finger \\markup
       \\concat {
         #start
         \\overtie \\concat { #left \\char ##x2009 #right } }
   #})
""",

    "for-barline": """\
% This `\\caesura` creates a light-light bar line where there isn't
% already a more significant bar line, and aligns its decorations
% (e.g., fermatas) to the bar line.
forBarLine =
{
  \\once \\set Staff.caesuraType = #'((underlying-bar-line . "||"))
  \\once \\set Staff.caesuraTypeTransform = ##f
  \\caesura
}
""",

    "harmonic": """\
harmonicSmall =
  \\tweak NoteHead.font-size #-3
  \\tweak NoteHead.duration-log #4
  \\etc

harmonicParen =
#(define-music-function (size offset music)
   (number? number? ly:music?)
   #{
     \\tweak Parentheses.font-size #size
     \\tweak Parentheses.Y-offset #offset
     \\parenthesize
     #music
   #})
""",

    "hide-note": """\
hideNote =
  \\tweak Dots.transparent ##t
  \\tweak NoteHead.transparent ##t
  \\tweak NoteHead.no-ledgers ##t
  \\tweak Stem.transparent ##t
  \\tweak Accidental.transparent ##t
  \\tweak Rest.transparent ##t
  \\tweak TabNoteHead.transparent ##t \\etc
""",

    "insert-before": """\
#(define (insert-before where what lst)
   (cond ((null? lst)
          (list what))
         ((eq? where (car lst))
          (cons what lst))
         (else
          (cons (car lst) (insert-before where what (cdr lst))))))
""",

    "make-bracketed": """\
#(define make-bracketed
   (grob-transformer
    'stencil
    (lambda (grob orig)
      (let* ((paren-stil (grob-interpret-markup
                          grob
                          (markup #:musicglyph "accidentals.leftparen")))
             (ext (ly:stencil-extent paren-stil Y))
             (stil (ly:accidental-interface::print grob))
             (thick (ly:output-def-lookup (ly:grob-layout grob)
                                          'line-thickness 0.1))
             (padding thick)
             (protrusion (* 2.5 thick))
             (lb (ly:bracket Y ext thick protrusion))
             (rb (ly:bracket Y ext thick (- protrusion))))
        (set! stil (ly:stencil-combine-at-edge stil X 1 rb padding))
        (set! stil (ly:stencil-combine-at-edge stil X -1 lb padding))
        stil))))

bracketAcc =
  \\tweak AccidentalCautionary.parenthesized ##f
  \\tweak AccidentalCautionary.stencil #make-bracketed \\etc
""",

    # TODO: Implement values `both` and `arrow` of `line-end` attribute.
    "make-edge-height": """\
#(define edge-height-alist
   '((up . -0.7)
     (down . 0.7)
     (none . 0)))

#(define (get-height type)
   (or (assv-ref edge-height-alist type)
       (begin
         (ly:warning "bracket edge type '~a' not implemented" type)
         0)))

% Make the `edge-height` property independent of a bracket's direction.
#(define (make-edge-height left right)
   (grob-transformer
    'edge-height
    (lambda (grob orig)
      (let ((dir (ly:grob-property grob 'direction))
            (left-height (get-height left))
            (right-height (get-height right)))
        (cons (* dir left-height) (* dir right-height))))))
""",

    "pluck": """\
RH = \\rightHandFinger \\etc
""",

    "square": """\
#(define-markup-command (square layout props arg)
   (markup?)
   (let* ((stil (interpret-markup layout props arg))
          (x-size (interval-length (ly:stencil-extent stil X)))
          (y-size (interval-length (ly:stencil-extent stil Y)))
          (axis (if (> x-size y-size) Y X)))
     (interpret-markup
      layout props
      #{
        \\markup \\box \\with-dimension-from #axis \\rotate #90 #arg #arg
      #})))
""",

    "staff-lines": """\
% Change the number of staff lines to `num-lines`.
%
% The argument `clef` makes the function emit a clef with the given
% name.  If `clef` contains the substring `"percussion"` or `"tab"`,
% use both even and odd staff line positions.  If set to `""`, do the
% same but don't set `Staff.clefPosition` (which means that no clef
% gets triggered).  If `clef` is set to any other value, use even
% staff line positions only and set `Staff.clefPosition`, which
% triggers a clef if its value changes.
%
% The optional argument `properties` is an alist of properties to
% control the appearance of both the staff and the clef:
%
% * `details` is a list of staff line numbers that should be
%   displayed.  An empty list suppresses any display of staff lines;
%   omitting the argument means to display all lines.
% * `staff-color` holds the color of the staff.
% * `clef-font-size` holds the font size of the clef
% * `clef-color` holds the color of the clef.
%
% \\staffLines [<properties>] <clef> <num-lines>

staffLines =
#(define-music-function (properties clef num-lines)
                        ((alist? '()) string? index?)
   (let* ((details (assoc-get 'details properties #f))
          (staff-color (assoc-get 'staff-color properties #f))
          (clef-color (assoc-get 'clef-color properties #f))
          (clef-font-size (assoc-get 'clef-font-size properties #f))
          (lines (or details (iota num-lines 1)))
          (only-even-pos (not (or (equal? clef "")
                                  (string-contains clef "percussion")
                                  (string-contains clef "tab"))))
          (offset (if only-even-pos
                      (if (even? num-lines) 1 0)
                      0))
          ;; MusicXML counts staff lines from the bottom.
          (delta (- -1 num-lines offset))
          (positions (filter (lambda (x) (<= x num-lines))
                             lines))
          (positions (map (lambda (x) (+ delta (* x 2)))
                      positions)))
     #{
       \\stopStaff

       #(if staff-color
            #{
              \\once \\override Staff.StaffSymbol.color = #staff-color
            #})
       #(if clef-color
            #{
              \\once \\override Staff.Clef.color = #clef-color
            #})
       #(if clef-font-size
            #{
              \\once \\override Staff.Clef.font-size = #clef-font-size
            #})

       #(if (equal? clef "")
            #{
              \\unset Staff.middleCPosition
              \\unset Staff.middleCClefPosition
              % To suppress the clef inserted by default by LilyPond at the
              % beginning of a piece.
              \\once \\omit Staff.Clef
            #}
            #{ \\clef #clef #})

       \\override Staff.StaffSymbol.line-positions = #positions

       \\applyContext
       #(lambda (ctx)
          (let* ((c (ly:context-find ctx 'Staff))
                 (clef-p (ly:context-property c 'clefPosition))
                 (middle-c-p (ly:context-property c 'middleCPosition))
                 (middle-c-clef-p (ly:context-property c 'middleCClefPosition))
                 (use-offset (not (string-contains clef "percussion")))
                 (delta (+ delta (if use-offset 6 0))))
            (when only-even-pos
              (ly:context-set-property! c 'clefPosition (+ clef-p delta))
              (ly:context-set-property! c 'forceClef #t))
            (ly:context-set-property! c 'middleCPosition (+ middle-c-p delta))
            (ly:context-set-property! c 'middleCClefPosition
                                      (+ middle-c-clef-p delta))))
       \\startStaff
     #}))
""",

    "stem-directions": """\
D = \\tweak Stem.direction #DOWN \\etc
U = \\tweak Stem.direction #UP \\etc
""",
}



# This file is part of LilyPond, the GNU music typesetter.
#
# Copyright (C) 2001--2026 Han-Wen Nienhuys <hanwen@xs4all.nl>
#           Jan Nieuwenhuizen <janneke@gnu.org>
#
#
# LilyPond is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# LilyPond is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with LilyPond.  If not, see <http://www.gnu.org/licenses/>.

# import midi
# s = open ("s.midi").read ()
# midi.parse_track (s)
# midi.parse (s)
#
#
# returns a MIDI file as the tuple
#
#  ((format, division), TRACKLIST)          # division (>0) = TPQN*4
#                                           # or (<0) TBD
#
# each track is an EVENTLIST, where EVENT is
#
#   (time, (type, ARG1, [ARG2]))            # time = cumulative delta time
# MIDI event:
#   type = MIDI status+channel >= x80
# META-event  = xFF:
#   type = meta-event type <= x7F
#   ARG1 = length
#   ARG2 = data

import array
import struct


class error (Exception):
    pass

# class warning (Exception): pass


# TODO: should use enum?

# Channel voice messages
NOTE_OFF = 0x80
NOTE_ON = 0x90
POLYPHONIC_KEY_PRESSURE = 0xa0
CONTROLLER_CHANGE = 0xb0
PROGRAM_CHANGE = 0xc0
CHANNEL_KEY_PRESSURE = 0xd0
PITCH_BEND = 0xe0

# Channel mode messages
ALL_SOUND_OFF = 0x78
RESET_ALL_CONTROLLERS = 0x79
LOCAL_CONTROL = 0x7a
ALL_NOTES_OFF = 0x7b
OMNI_MODE_OFF = 0x7c
OMNI_MODE_ON = 0x7d
MONO_MODE_ON = 0x7e
POLY_MODE_ON = 0x7f

# Meta events
SEQUENCE_NUMBER = 0x0
TEXT_EVENT = 0x1
COPYRIGHT_NOTICE = 0x2
SEQUENCE_TRACK_NAME = 0x3
INSTRUMENT_NAME = 0x4
LYRIC = 0x5
MARKER = 0x6
CUE_POINT = 0x7
PROGRAM_NAME = 0x8
DEVICE_NAME = 0x9
MIDI_CHANNEL_PREFIX = 0x20
MIDI_PORT = 0x21
END_OF_TRACK = 0x2f
SET_TEMPO = 0x51
SMTPE_OFFSET = 0x54
TIME_SIGNATURE = 0x58
KEY_SIGNATURE = 0x59
XMF_PATCH_TYPE_PREFIX = 0x60
SEQUENCER_SPECIFIC_META_EVENT = 0x7f
META_EVENT = 0xff


def _get_variable_length_number(nextbyte, getbyte):
    sum = 0
    while nextbyte >= 0x80:
        sum = (sum + (nextbyte & 0x7F)) << 7
        nextbyte = getbyte()
    return sum + nextbyte


def _first_command_is_repeat(status, nextbyte, getbyte):
    raise error('the first midi command in the track is a repeat')


def _read_two_bytes(status, nextbyte, getbyte):
    return status, nextbyte


def _read_three_bytes(status, nextbyte, getbyte):
    return status, nextbyte, getbyte()


def _read_string(nextbyte, getbyte):
    length = _get_variable_length_number(nextbyte, getbyte)
    return ''.join(chr(getbyte()) for i in range(length))


def _read_f0_byte(status, nextbyte, getbyte):
    if status == 0xff:
        return status, nextbyte, _read_string(getbyte(), getbyte)
    return status, _read_string(nextbyte, getbyte)


_read_midi_event = (
    _first_command_is_repeat,  # 0
    None,  # 10
    None,  # 20
    None,  # 30
    None,  # 40
    None,  # 50
    None,  # 60 data entry???
    None,  # 70 all notes off???
    _read_three_bytes,  # 80 note off
    _read_three_bytes,  # 90 note on
    _read_three_bytes,  # a0 poly aftertouch
    _read_three_bytes,  # b0 control
    _read_two_bytes,  # c0 prog change
    _read_two_bytes,  # d0 ch aftertouch
    _read_three_bytes,  # e0 pitchwheel range
    _read_f0_byte,   # f0
)


def _parse_track_body(data, clocks_max):
    # This seems to be the fastest way of getting bytes in order as integers.
    dataiter = iter(array.array('B', data))
    getbyte = dataiter.__next__

    time = 0
    status = 0
    try:
        for nextbyte in dataiter:
            time += _get_variable_length_number(nextbyte, getbyte)
            if clocks_max and time > clocks_max:
                break
            nextbyte = getbyte()
            if nextbyte >= 0x80:
                status = nextbyte
                nextbyte = getbyte()
            yield time, _read_midi_event[status >> 4](status, nextbyte, getbyte)
    except StopIteration:
        # If the track ended just before the start of an event, the for loop
        # will exit normally. If it ends anywhere else, we end up here.
        print(len(list(dataiter)))
        raise error('a track ended in the middle of a MIDI command')


def _parse_hunk(data, pos, type, magic):
    if data[pos:pos+4] != magic:
        raise error('expected %r, got %r' % (magic, data[pos:pos+4]))
    try:
        length, = struct.unpack('>I', data[pos+4:pos+8])
    except struct.error:
        raise error(
            'the %s header is truncated (may be an incomplete download)' % type)
    endpos = pos + 8 + length
    data = data[pos+8:endpos]
    if len(data) != length:
        raise error(
            'the %s is truncated (may be an incomplete download)' % type)
    return data, endpos


def _parse_tracks(midi, pos, num_tracks, clocks_max):
    if num_tracks > 256:
        raise error('too many tracks: %d' % num_tracks)
    for i in range(num_tracks):
        trackdata, pos = _parse_hunk(midi, pos, 'track', b'MTrk')
        yield list(_parse_track_body(trackdata, clocks_max))
    # if pos < len(midi):
    #     warn


def parse_track(track, clocks_max=None):
    track_body, end = _parse_hunk(track, 0, 'track', b'MTrk')
    # if end < len(track):
    #     warn
    return list(_parse_track_body(track_body, clocks_max))


def parse(midi, clocks_max=None):
    header, first_track_pos = _parse_hunk(midi, 0, 'file', b'MThd')
    try:
        format, num_tracks, division = struct.unpack('>3H', header[:6])
    except struct.error:
        raise error('the file header is too short')
#  if division < 0:
#    raise error ('cannot handle non-metrical time')
    tracks = list(_parse_tracks(midi, first_track_pos, num_tracks, clocks_max))
    return (format, division*4), tracks
